@@ -23,11 +23,16 @@ __all__ = [
     "Literal", "Word", "Repetition", "ListRepetition", "Reference",
     "GRAMMAR", "G", "ANY", "EMPTY", "REF", "LITERAL", "L", "OR", "EXCEPT", "WORD", "REPEAT", "LIST_OF", "OPTIONAL", "NOT_FOLLOWED_BY",
     "ZERO_OR_MORE", "ONE_OR_MORE", "ANY_EXCEPT", "BOL", "EOL", "EOF",
-    "REST_OF_LINE", "SPACE",
+    "REST_OF_LINE", "WHITESPACE", "SPACE",
     "generate_ebnf",
+    "WS_DEFAULT", "WS_NOEOL",
 ]
 
-grammar_whitespace = True
+WS_DEFAULT = re.compile(r'\s+')
+WS_NOEOL = re.compile(r'[^\S\r\n]+')
+
+grammar_whitespace = WS_DEFAULT
+grammar_whitespace_mode = 'optional'
 
 class _Singleton:
   def __init__(self, name):
@@ -38,8 +43,6 @@ class _Singleton:
 
 DEFAULT = _Singleton("DEFAULT") # singleton used for detecting default arguments
 
-PARSEERROR_FOUNDTXT_LEN = 16
-
 def _gclass_reconstructor(name, bases, cdict):
   return GrammarClass(name, bases, cdict)
 
@@ -124,12 +127,7 @@ class ParseError (Exception):
         message = ""
       else:
         expected_txt = " or ".join(sorted(e.grammar_desc for e in expected))
-        end = min(len(buf), pos + PARSEERROR_FOUNDTXT_LEN)
-        found_txt = buf[pos:end]
-        if found_txt:
-          found_txt = repr(found_txt)
-        else:
-          found_txt = "(end of input)"
+        found_txt = util.get_found_txt(buf, pos)
         message = "Expected {}: Found {}".format(expected_txt, found_txt)
     self.buffer = buf
     self.buffer_pos = pos
@@ -158,22 +156,38 @@ class GrammarClass (type):
   "The metaclass for all Grammar classes"
 
   def __init__(cls, name, bases, classdict):
+    if getattr(cls, 'grammar', None) is None:
+      # This is an abstract class definition.  Don't do any of our usual setup.
+      return
     cls._hash_id = None
     if "grammar_name" not in classdict:
       cls.grammar_name = cls.__name__
     if "grammar_desc" not in classdict:
       cls.grammar_desc = cls.grammar_name
-    classdict["abstract"] = False
     cls.grammar = util.regularize(cls.grammar)
+    if not hasattr(cls, 'grammar_min'):
+      cls.grammar_min = len(cls.grammar)
+    if not hasattr(cls, 'grammar_max'):
+      cls.grammar_max = len(cls.grammar)
     tags = getattr(cls, "grammar_tags", ())
     if isinstance(tags, str):
       # This is going to be a common slip-up.. might as well handle it
       # gracefully...
       tags = (tags,)
     cls.grammar_tags = tags
+    mod = sys.modules[cls.__module__]
     if "grammar_whitespace" not in classdict and cls.grammar_whitespace is None:
-      whitespace = sys.modules[cls.__module__].__dict__.get("grammar_whitespace", grammar_whitespace)
+      whitespace = getattr(mod, "grammar_whitespace", grammar_whitespace)
       cls.grammar_whitespace = whitespace
+    if cls.grammar_whitespace is False:
+      util.depwarning("grammar_whitespace=False is deprecated.  Use grammar_whitespace_mode='explicit' instead.", module=mod)
+    elif cls.grammar_whitespace is True:
+      util.depwarning("grammar_whitespace=True is deprecated.  Use grammar_whitespace_mode='optional' instead.", module=mod)
+    if "grammar_whitespace_mode" not in classdict and cls.grammar_whitespace_mode is None:
+      if not hasattr(mod, "grammar_whitespace") and not hasattr(mod, "grammar_whitespace_mode"):
+        util.depwarning("default whitespace mode will be changing.  For future compatibility, set grammar_whitespace_mode='optional' explicitly.", module=mod)
+      whitespace_mode = getattr(mod, "grammar_whitespace_mode", grammar_whitespace_mode)
+      cls.grammar_whitespace_mode = whitespace_mode
     cls.__class_init__(classdict)
 
   def __reduce__(cls):
@@ -401,7 +415,7 @@ class GrammarParser:
       result = result[0]
     return (count, result)
 
-  def _parse_string(self, string, bol, eof, data, matchtype):
+  def _parse_text(self, string, bol, eof, data, matchtype):
     self.append(string, bol=bol, eof=eof)
     pos = 0
     if data is None:
@@ -426,7 +440,7 @@ class GrammarParser:
 	# case or the count-is-zero case next time through.
         break
 
-  def parse_string(self, string, bol=None, eof=None, reset=False, multi=False, data=None, matchtype='first'):
+  def parse_text(self, string, bol=None, eof=None, reset=False, multi=False, data=None, matchtype='first'):
     """
     Attempt to match *string* against the associated grammar.  If successful, returns a corresponding match object.  If there is an incomplete match (or it is impossible to determine yet whether the match is complete or not), save the current text in the match buffer and return :const:`None` to indicate more text is required.  If the text does not match any valid grammar construction, raise :exc:`ParseError`.
 
@@ -457,31 +471,35 @@ class GrammarParser:
     if reset:
       self.reset()
     if multi:
-      return list(self._parse_string(string, bol, eof, data, matchtype))
+      return list(self._parse_text(string, bol, eof, data, matchtype))
     else:
-      for result in self._parse_string(string, bol, eof, data, matchtype):
+      for result in self._parse_text(string, bol, eof, data, matchtype):
         # This will always just return the first result
         return result
       return None
 
+  def parse_string(self, *args, **kwargs):
+    util.depwarning("parse_string syntax will be changing: For future compatibility, use parse_text instead.")
+    return self.parse_text(*args, **kwargs)
+
   def parse_lines(self, lines, bol=False, eof=False, reset=False, data=None, matchtype='first'):
     """
     *(generator method)*
 
-    Attempt to match a list (or actually any iterable) of strings against the associated grammar.  This is effectively the same as calling :meth:`parse_string` repeatedly for each string in the list to obtain all matches in sequence.
+    Attempt to match a list (or actually any iterable) of strings against the associated grammar.  This is effectively the same as calling :meth:`parse_text` repeatedly for each string in the list to obtain all matches in sequence.
 
-    Return values, exceptions, and optional parameters are all exactly the same as for :meth:`parse_string`.
+    Return values, exceptions, and optional parameters are all exactly the same as for :meth:`parse_text`.
 
-    Note: Be careful using ``matchtype="all"`` with parse_lines/parse_file.  You must manually call :func:`~GrammarParser.skip` after each yielded match, or you will end up with an infinite loop!
+    **Note:** Be careful using ``matchtype="all"`` with parse_lines/parse_file.  You must manually call :func:`~GrammarParser.skip` after each yielded match, or you will end up with an infinite loop!
     """
     if reset:
       self.reset()
     for line in lines:
-      for result in self._parse_string(line, bol, False, data, matchtype):
+      for result in self._parse_text(line, bol, False, data, matchtype):
         yield result
       bol = None
     if eof:
-      for result in self._parse_string("", None, True, data, matchtype):
+      for result in self._parse_text("", None, True, data, matchtype):
         yield result
 
   def parse_file(self, file, bol=False, eof=True, reset=False, data=None, matchtype='first'):
@@ -490,9 +508,9 @@ class GrammarParser:
 
     Open and process the contents of a file using the associated grammar.  This is basically the same as opening the specified file, and passing the resulting file object to :meth:`parse_lines`.
 
-    Return values, exceptions, and optional parameters are all exactly the same as for :meth:`parse_string`.
+    Return values, exceptions, and optional parameters are all exactly the same as for :meth:`parse_text`.
 
-    Note: Be careful using ``matchtype="all"`` with parse_lines/parse_file.  You must manually call :func:`~GrammarParser.skip` after each yielded match, or you will end up with an infinite loop!
+    **Note:** Be careful using ``matchtype="all"`` with parse_lines/parse_file.  You must manually call :func:`~GrammarParser.skip` after each yielded match, or you will end up with an infinite loop!
     """
     if isinstance(file, str):
       with open(file, "r") as f:
@@ -536,19 +554,18 @@ class Grammar (metaclass=GrammarClass):
   To define a new grammar, you should create a new class definition, descended from :class:`Grammar`.  In this class definition, you can override several class attributes and class methods to customize the behavior of the grammar.
   """
 
-  grammar = ()
   grammar_terminal = False
   grammar_collapse = False
   grammar_greedy = True
   grammar_null_subtoken_ok = True
   grammar_whitespace = None
+  grammar_whitespace_mode = None
   grammar_error_override = False
-  grammar_hashattrs = ('grammar_name', 'grammar', 'grammar_min', 'grammar_max', 'grammar_collapse', 'grammar_greedy', 'grammar_whitespace')
+  grammar_hashattrs = ('grammar_name', 'grammar', 'grammar_min', 'grammar_max', 'grammar_collapse', 'grammar_greedy', 'grammar_whitespace', 'grammar_whitespace_mode')
 
   @classmethod
   def __class_init__(cls, attrs):
-    cls.grammar_min = len(cls.grammar)
-    cls.grammar_max = len(cls.grammar)
+    pass
 
   @classmethod
   def parser(cls, sessiondata=None, tabs=1):
@@ -577,10 +594,18 @@ class Grammar (metaclass=GrammarClass):
     grammar_min = cls.grammar_min
     grammar_max = cls.grammar_max
     greedy = cls.grammar_greedy
-    if cls.grammar_whitespace is True:
-      whitespace_re = util._whitespace_re
-    else:
+    whitespace_mode = cls.grammar_whitespace_mode
+    if whitespace_mode == 'optional':
       whitespace_re = cls.grammar_whitespace
+      whitespace_reqd = False
+    elif whitespace_mode == 'required':
+      whitespace_re = cls.grammar_whitespace
+      whitespace_reqd = True
+    else:
+      whitespace_re = False
+      whitespace_reqd = False
+    if whitespace_re is True:
+      whitespace_re = util._whitespace_re
     objs = []
     states = []
     positions = []
@@ -611,6 +636,14 @@ class Grammar (metaclass=GrammarClass):
             if pos < len(text.string) or text.eof:
               break
             text = yield (None, None)
+          if whitespace_reqd and objs and pos == prews_pos:
+	    # We didn't match any whitespace before the next sub-grammar, but
+	    # whitespace is required between sub-grammars.  Handle this as if
+	    # there were a WHITESPACE grammar in this spot that gave us back an
+	    # error result.
+            obj = util.error_result(pos, WHITESPACE)[1]
+            best_error = util.update_best_error(best_error, obj)
+            break
         if first_pos is None:
           first_pos = pos
         s = grammar[len(objs)].grammar_parse(text, pos, sessiondata)
@@ -733,7 +766,7 @@ class Grammar (metaclass=GrammarClass):
   @classmethod
   def __class_str__(cls):
     """
-    Returns the string to be used when :func:`str` is used on this grammar class (Note: This is for the class itself, not for instances of the class.  For those, the usual :meth:`__str__` is used).
+    Returns the string to be used when :func:`str` is used on this grammar class (**Note:** This is for the class itself, not for instances of the class.  For those, the usual :meth:`__str__` is used).
     """
 
     return cls.grammar_name
@@ -741,7 +774,7 @@ class Grammar (metaclass=GrammarClass):
   @classmethod
   def __class_repr__(cls):
     """
-    Returns the string to be used when :func:`repr` is used on this grammar class (Note: This is for the class itself, not for instances of the class.  For those, the usual :meth:`__repr__` is used).
+    Returns the string to be used when :func:`repr` is used on this grammar class (**Note:** This is for the class itself, not for instances of the class.  For those, the usual :meth:`__repr__` is used).
     """
 
     name = cls.grammar_name
@@ -852,41 +885,43 @@ class Grammar (metaclass=GrammarClass):
     """
     pass
 
-  def get_all(self, *type_path):
+  def get_all(self, *tt_path):
     """
-    Return all immediate sub-elements of the given type.
+    Return all immediate sub-elements of the given type, or posessing the given tag.
 
-    If more than one type parameter is provided, it will treat the types as a "path" to traverse: for all sub-elements matching the first type, retrieve all sub-elements of those matching the second type, and so on, until it reaches the last type in the list.  (It will thus return all elements of the parse tree which are of the final type, which can be reached by traversing the previous types in order.)
+    If more than one type/tag parameter is provided, it will treat the types as a "path" to traverse: for all sub-elements matching the first type/tag, retrieve all sub-elements of those matching the second type/tag, and so on, until it reaches the last type/tag in the list.  (It will thus return all elements of the parse tree which are of the final type/tag, which can be reached by traversing the previous types/tags in order.)
+
+    Note that types and tags can be intermingled, as well.
     """
-    return list(self._search_recursive(isinstance, False, type_path))
+    return list(self._search_recursive(util.find_match_func, False, tt_path))
 
-  def get(self, *type_path):
+  def get(self, *tt_path):
     """
-    Return the first immediate sub-element of the given type (or by descending through multiple types, in the same way as :meth:`get_all`).
+    Return the first immediate sub-element of the given type/tag (or by descending through multiple types/tags, in the same way as :meth:`get_all`).
 
-    This is equivalent to ``.get_all(*type_path)[0]`` except that it is more efficient, and will return :const:`None` if there are no such objects (instead of raising :exc:`IndexError`).
+    This is equivalent to ``.get_all(*tt_path)[0]`` except that it is more efficient, and will return :const:`None` if there are no such objects (instead of raising :exc:`IndexError`).
     """
     try:
-      return next(self._search_recursive(isinstance, False, type_path))
+      return next(self._search_recursive(util.find_match_func, False, tt_path))
     except StopIteration:
       return None
 
-  def find_all(self, *type_path):
+  def find_all(self, *tt_path):
     """
-    Return all elements anywhere in the parse tree matching the given type.
+    Return all elements anywhere in the parse tree matching the given type/tag.
 
-    Similar to :meth:`get_all`, if more than one type parameter is provided, it will treat the types as a "path" to traverse in order.  The difference from :meth:`get_all` is that, for each step in the path, the elements found do not have to be direct sub-elements, but can be anywhere in the sub-tree.
+    Similar to :meth:`get_all`, if more than one type/tag parameter is provided, it will treat them as a "path" to traverse in order.  The difference from :meth:`get_all` is that, for each step in the path, the elements found do not have to be direct sub-elements, but can be anywhere in the sub-tree.
     """
-    return list(self._search_recursive(isinstance, True, type_path))
+    return list(self._search_recursive(util.find_match_func, True, tt_path))
 
-  def find(self, *type_path):
+  def find(self, *tt_path):
     """
-    Return the first element anywhere in the parse tree matching the given type (or by descending through multiple types, in the same way as :meth:`find_all`).
+    Return the first element anywhere in the parse tree matching the given type/tag (or by descending through multiple types/tags, in the same way as :meth:`find_all`).
 
-    This is equivalent to ``.find_all(*type_path)[0]`` except that it is more efficient, and will return :const:`None` if there are no such objects (instead of raising :exc:`IndexError`).
+    This is equivalent to ``.find_all(*tt_path)[0]`` except that it is more efficient, and will return :const:`None` if there are no such objects (instead of raising :exc:`IndexError`).
     """
     try:
-      return next(self._search_recursive(isinstance, True, type_path))
+      return next(self._search_recursive(util.find_match_func, True, tt_path))
     except StopIteration:
       return None
 
@@ -894,8 +929,9 @@ class Grammar (metaclass=GrammarClass):
     """
     Return all elements anywhere in the parse tree with the given tag.
 
-    This functions identically to :meth:`find_all`, except that the criteria for matching is based on tags, rather than object types.
+    **Note:** This method is deprecated.  Its functionality is now part of :meth:`find_all` instead.
     """
+    util.depwarning("find_tag_all is deprecated: Use find_all instead.")
     func = lambda e, l: l in getattr(e, "grammar_tags", ())
     return list(self._search_recursive(func, True, tag_path))
 
@@ -903,8 +939,9 @@ class Grammar (metaclass=GrammarClass):
     """
     Return the first element anywhere in the parse tree with the given tag (or by descending through multiple tags, in the same way as :meth:`find_tag_all`).
 
-    This is equivalent to ``.find_tag_all(*tag_path)[0]`` except that it is more efficient, and will return :const:`None` if there are no such objects (instead of raising :exc:`IndexError`).
+    **Note:** This method is deprecated.  Its functionality is now part of :meth:`find` instead.
     """
+    util.depwarning("find_tag is deprecated: Use find instead.")
     func = lambda e, l: l in getattr(e, "grammar_tags", ())
     try:
       return next(self._search_recursive(func, True, tag_path))
@@ -946,7 +983,13 @@ class Grammar (metaclass=GrammarClass):
     return [e.string for e in self.terminals()]
 
   def __getitem__(self, index):
-    return self.elements[index]
+    if isinstance(index, int):
+      return self.elements[index]
+    else:
+      e = self.get(index)
+      if e is not None:
+        return e
+      raise IndexError('no subgrammar matching {!r} found'.format(index))
 
   def __len__(self):
     return len(self.string)
@@ -986,6 +1029,7 @@ class Grammar (metaclass=GrammarClass):
 
 class AnonGrammar (Grammar):
   grammar_whitespace = None
+  grammar_whitespace_mode = None
 
   @classmethod
   def grammar_details(cls, depth=-1, visited=None):
@@ -1011,15 +1055,16 @@ class AnonGrammar (Grammar):
       return (", ".join(names), nts)
 
 class Terminal (Grammar):
-  grammar = ()
   grammar_terminal = True
-  grammar_whitespace = False
 
   @classmethod
   def grammar_details(cls, depth=-1, visited=None):
     return cls.grammar_name
 
 class Literal (Terminal):
+  grammar_whitespace_mode = 'explicit'
+  grammar_whitespace = None
+  grammar = ()
   string = ""
   grammar_collapse_skip = True
   grammar_hashattrs = ('string',)
@@ -1074,6 +1119,9 @@ def LITERAL(string, **kwargs):
   return GrammarClass("<LITERAL>", (Literal,), cdict)
 
 class ANY (Terminal):
+  grammar_whitespace_mode = 'explicit'
+  grammar_whitespace = None
+  grammar = ()
   grammar_name = "ANY"
   grammar_desc = "any character"
 
@@ -1088,6 +1136,9 @@ class ANY (Terminal):
     yield error_result(index, cls)
 
 class EMPTY (Terminal):
+  grammar_whitespace_mode = 'explicit'
+  grammar_whitespace = None
+  grammar = ()
   grammar_collapse = True
   grammar_collapse_skip = True
   grammar_desc = "(nothing)"
@@ -1110,7 +1161,7 @@ def OR(*grammars, **kwargs):
   """
   An either-or grammar that will successfully match if any of its subgrammars matches.  :func:`OR` grammars can also be created by combining other grammars in python expressions using the or operator (``|``).
 
-  Note: Each of the possible grammars are attempted in left-to-right order.  This means that if more than one of the listed grammars could potentially match, the leftmost one will always match first.
+  **Note:** Each of the possible grammars are attempted in left-to-right order.  This means that if more than one of the listed grammars could potentially match, the leftmost one will always match first.
   """
   collapsed = []
   for g in grammars:
@@ -1122,7 +1173,7 @@ def OR(*grammars, **kwargs):
   return GrammarClass("<OR>", (OR_Operator,), cdict)
 
 class OR_Operator (Grammar):
-  grammar_whitespace = False
+  grammar_whitespace_mode = 'explicit'
 
   @classmethod
   def __class_init__(cls, attrs):
@@ -1181,7 +1232,7 @@ def NOT_FOLLOWED_BY(*grammar, **kwargs):
   return GrammarClass("<NOT_FOLLOWED_BY>", (NotFollowedBy,), cdict)
 
 class NotFollowedBy (Grammar):
-  grammar_whitespace = False
+  grammar_whitespace_mode = 'explicit'
   grammar_collapse = True
 
   @classmethod
@@ -1212,6 +1263,9 @@ class NotFollowedBy (Grammar):
     else:
       # Subgrammar did not match.  Return a (successful) None match.
       yield (0, cls(''))
+      # If the caller proceeds to the next "match", we need to yield an error
+      # to indicate we're done with matching.
+      yield error_result(index, cls)
     
   @classmethod
   def grammar_details(cls, depth=-1, visited=None):
@@ -1238,13 +1292,13 @@ def EXCEPT(grammar, exc_grammar, **kwargs):
   """
   Match *grammar*, but only if it does not also match *exception_grammar*.  (This is equivalent to the ``-`` (exception) operator in EBNF) :func:`EXCEPT` grammars can also be created by combining other grammars in python expressions using the except operator (``-``).
 
-  Note: In many cases there are more efficient ways to design a particular grammar than using this construct.  It is provided mostly for full EBNF compatibility.
+  **Note:** In many cases there are more efficient ways to design a particular grammar than using this construct.  It is provided mostly for full EBNF compatibility.
   """
   cdict = util.make_classdict(ExceptionGrammar, (grammar, exc_grammar), kwargs)
   return GrammarClass("<EXCEPT>", (ExceptionGrammar,), cdict)
 
 class ExceptionGrammar (Grammar):
-  grammar_whitespace = False
+  grammar_whitespace_mode = 'explicit'
 
   @classmethod
   def __class_init__(cls, attrs):
@@ -1275,7 +1329,7 @@ class ExceptionGrammar (Grammar):
         if e_count is None:
           # Subgrammars should not be asking for more data after eof.
           raise InternalError("{} requested more data when at EOF".format(g))
-        if e_count is False:
+        if e_count != count:
           break
         found = True
         break
@@ -1284,7 +1338,7 @@ class ExceptionGrammar (Grammar):
     # In some cases, our "best error" can lead to really confusing messages,
     # since it may say "expected foo" at a place where foo actually WAS found
     # (because the exclusion grammar prevented it from being returned).  If
-    # this is the case (we"re returning a best error at our own starting
+    # this is the case (we're returning a best error at our own starting
     # position) return ourselves as the error object, so at least it will be
     # obvious there were extra conditions on the match that weren't fulfilled.
     if best_error[0] == index:
@@ -1321,6 +1375,7 @@ class Repetition (Grammar):
   grammar_min = 1
   grammar_max = None
   grammar_whitespace = None
+  grammar_whitespace_mode = None
   
   @classmethod
   def __class_init__(cls, attrs):
@@ -1390,18 +1445,23 @@ class Repetition (Grammar):
   def grammar_ebnf_rhs(cls, opts):
     return None
 
-def WORD(startchars, restchars=None, **kwargs):
+def WORD(startchars, restchars=None, longest=False, **kwargs):
   """
   Match any text consisting of a sequence of the specified characters.  If *restchars* is not provided, all characters in the sequence must be in the set specified by *startchars*.  If *restchars* is provided, then *startchars* specifies the valid options for the first character of the sequence, and *restchars* specifies the valid options for all following characters.
 
   *startchars* and *restchars* are each strings containing a sequence of individual characters, or character ranges, in the same format used by python regular expressions for character-range (``[]``) operations (i.e. ``"0123456789"`` or ``"A-Za-z"``).  If the first character of *startchars* or *restchars* is ``^``, the meaning is also inverted, just as in regular expressions, so ``"^A-Z"`` would match anything *except* an upper-case ascii alphabet character.
+
+  If *longest* is true, then this construct will only match the longest possible sequence of matching characters in the input, and will not try to match any shorter sub-sequences (even if they would work better in the context of the larger grammar).  Effectively, this means any match returned is guaranteed to be followed by a non-word character or token.
   """
-  cdict = util.make_classdict(Word, (), kwargs, startchars=startchars, restchars=restchars)
+  cdict = util.make_classdict(Word, (), kwargs, startchars=startchars, restchars=restchars, longest_only=longest)
   return GrammarClass("<WORD>", (Word,), cdict)
 
 class Word (Terminal):
+  grammar_whitespace_mode = 'explicit'
+  grammar = ()
   startchars = ""
   restchars = None
+  longest_only = False
   grammar_count = None
   grammar_min = 1
   grammar_max = None
@@ -1448,11 +1508,13 @@ class Word (Terminal):
       argspec += ", min={}".format(min)
     if max:
       argspec += ", max={}".format(max)
+    if cls.longest_only:
+      argspec += ", longest=True"
     return "WORD({})".format(argspec)
 
   @classmethod
   def grammar_parse(cls, text, index, sessiondata):
-    greedy = cls.grammar_greedy
+    greedy = cls.grammar_greedy or cls.longest_only
     returned = cls.grammar_min - 1
     while True:
       string = text.string
@@ -1472,6 +1534,8 @@ class Word (Terminal):
     if greedy:
       while matchlen >= cls.grammar_min:
         yield (matchlen, cls(string, index, index+matchlen))
+        if cls.longest_only:
+          break
         matchlen -= 1
     yield error_result(index, cls)
 
@@ -1506,7 +1570,7 @@ class Reference (Grammar):
   ref_name = None
   ref_base = None
   ref_default = None
-  grammar_whitespace = False
+  grammar_whitespace_mode = 'explicit'
 
   @classmethod
   def __class_init__(cls, attrs):
@@ -1572,7 +1636,7 @@ def LIST_OF(*grammar, **kwargs):
   """
   Match a list consisting of repetitions of *grammar* separated by *sep*.  As with other repetition grammars, the *min* and *max* keywords can also be used to restrict the number of matches to a certain range.
 
-  Note: Although this is most commonly used with a literal separator (such as the default ``","``), actually any (arbitrarily-complex) subgrammar can be specified for *sep* if desired.
+  **Note:** Although this is most commonly used with a literal separator (such as the default ``","``), actually any (arbitrarily-complex) subgrammar can be specified for *sep* if desired.
   """
   cdict = util.make_classdict(ListRepetition, grammar, kwargs)
   return GrammarClass("<LIST>", (ListRepetition,), cdict)
@@ -1581,13 +1645,14 @@ class ListRepetition (Repetition):
   sep = LITERAL(",")
   grammar_min = 1
   grammar_whitespace = None
+  grammar_whitespace_mode = None
 
   @classmethod
   def __class_init__(cls, attrs):
     grammar = GRAMMAR(cls.grammar)
     Repetition.__class_init__.__func__(cls, attrs)
     cls.sep = GRAMMAR(cls.sep)
-    succ_grammar = GRAMMAR(cls.sep, grammar, whitespace=cls.grammar_whitespace)
+    succ_grammar = GRAMMAR(cls.sep, grammar, whitespace=cls.grammar_whitespace, whitespace_mode=cls.grammar_whitespace_mode)
     cls.grammar = util.RepeatingTuple(grammar, succ_grammar, len=cls.grammar_max)
 
   @classmethod
@@ -1628,7 +1693,7 @@ def OPTIONAL(*grammar, **kwargs):
   kwargs.update(min=0, max=1)
   kwargs.setdefault("collapse", True)
   kwargs.setdefault("grammar_name", "<OPTIONAL>")
-  kwargs.setdefault("whitespace", False)
+  kwargs.setdefault("whitespace_mode", 'explicit')
   return REPEAT(*grammar, **kwargs)
 
 def ZERO_OR_MORE(*grammar, **kwargs):
@@ -1656,6 +1721,9 @@ def ANY_EXCEPT(charlist, **kwargs):
 
 # FIXME: whitespace at beginning of line
 class BOL (Terminal):
+  grammar_whitespace_mode = 'explicit'
+  grammar_whitespace = None
+  grammar = ()
   grammar_desc = "beginning of line"
 
   @classmethod
@@ -1668,6 +1736,9 @@ class BOL (Terminal):
     yield error_result(index, cls)
 
 class EOF (Terminal):
+  grammar_whitespace_mode = 'explicit'
+  grammar_whitespace = None
+  grammar = ()
   grammar_desc = "end of file"
 
   @classmethod
@@ -1677,10 +1748,25 @@ class EOF (Terminal):
     yield error_result(index, cls)
 
 class EOL (Terminal):
+  grammar_whitespace_mode = 'explicit'
+  grammar_whitespace = None
   grammar_desc = "end of line"
   grammar_collapse_skip = True
   grammar = (L("\n\r") | L("\r\n") | L("\r") | L("\n"))
 
+class WHITESPACE (Word):
+  grammar_desc = "whitespace"
+  regexp = re.compile("[\s]+")
+
+  @classmethod
+  def __class_init__(cls, attrs):
+    # Don't do the normal Word __class_init__ stuff.
+    pass
+
+  @classmethod
+  def grammar_details(cls, depth=-1, visited=None):
+    return cls.grammar_name
+
 class SPACE (Word):
   grammar_desc = "whitespace"
   regexp = re.compile("[\s]+")
@@ -1690,6 +1776,22 @@ class SPACE (Word):
     # Don't do the normal Word __class_init__ stuff.
     pass
 
+  @classmethod
+  def grammar_parse(cls, text, index, sessiondata):
+    util.depwarning("The meaning of SPACE will be changing: For future compatibility, use WHITESPACE instead.", util.get_calling_stacklevel() or 3)
+    s = Word.grammar_parse.__func__(cls, text, index, sessiondata)
+    offset, obj = next(s)
+    try:
+      while True:
+        if offset is None:
+          text = yield (offset, obj)
+          offset, obj = s.send(text)
+        else:
+          yield (offset, obj)
+          offset, obj = next(s)
+    except StopIteration:
+      pass
+
   @classmethod
   def grammar_details(cls, depth=-1, visited=None):
     return cls.grammar_name
