@@ -1,9 +1,12 @@
 import re
 import traceback
 import sys
+import warnings
 
 import modgrammar
 
+PARSEERROR_FOUNDTXT_LEN = 16
+
 _whitespace_re = re.compile('\s+')
 
 def update_best_error(current_best, err):
@@ -47,7 +50,7 @@ def regularize(grammar):
   except TypeError:
     raise modgrammar.GrammarDefError("object of type '%s' cannot be converted to Grammar" % (type(grammar).__name__,))
 
-_anongrammar_attrs = ('grammar_collapse', 'grammar_desc', 'grammar_name', 'grammar_whitespace', 'grammar_tags')
+_anongrammar_attrs = ('grammar_collapse', 'grammar_desc', 'grammar_name', 'grammar_whitespace', 'grammar_whitespace_mode', 'grammar_tags')
 
 def is_simple_anongrammar(cls):
   if not issubclass(cls, AnonGrammar):
@@ -66,7 +69,17 @@ def add_grammar(one, two):
     two = two[0].grammar
   return modgrammar.GRAMMAR(*(one + two))
 
-classdict_map = dict(count='grammar_count', min='grammar_min', max='grammar_max', collapse='grammar_collapse', collapse_skip='grammar_collapse_skip', tags='grammar_tags', greedy='grammar_greedy', whitespace='grammar_whitespace')
+classdict_map = dict(
+    count='grammar_count',
+    min='grammar_min',
+    max='grammar_max',
+    collapse='grammar_collapse',
+    collapse_skip='grammar_collapse_skip',
+    tags='grammar_tags',
+    greedy='grammar_greedy',
+    whitespace='grammar_whitespace',
+    whitespace_mode='grammar_whitespace_mode',
+)
 
 def make_classdict(base, grammar, kwargs, **defaults):
   cdict = {}
@@ -76,9 +89,21 @@ def make_classdict(base, grammar, kwargs, **defaults):
       cdict[key] = value
   cdict['grammar'] = grammar
   if not "grammar_whitespace" in cdict and base.grammar_whitespace is None:
-    mdict = get_calling_module().__dict__
-    whitespace = mdict.get("grammar_whitespace", modgrammar.grammar_whitespace)
+    mod = get_calling_module()
+    whitespace = getattr(mod, "grammar_whitespace", modgrammar.grammar_whitespace)
     cdict["grammar_whitespace"] = whitespace
+  if cdict.get("grammar_whitespace", None) is False:
+    mod = get_calling_module()
+    depwarning("grammar_whitespace=False is deprecated.  Use grammar_whitespace_mode='explicit' instead.", get_calling_stacklevel() or 3, mod)
+  elif cdict.get("grammar_whitespace", None) is True:
+    mod = get_calling_module()
+    depwarning("grammar_whitespace=True is deprecated.  Use grammar_whitespace_mode='optional' instead.", get_calling_stacklevel() or 3, mod)
+  if not "grammar_whitespace_mode" in cdict and base.grammar_whitespace_mode is None:
+    mod = get_calling_module()
+    if not hasattr(mod, "grammar_whitespace") and not hasattr(mod, "grammar_whitespace_mode"):
+      depwarning("default whitespace mode will be changing.  For future compatibility, set grammar_whitespace_mode='optional' explicitly.", get_calling_stacklevel() or 3, mod)
+    whitespace_mode = getattr(mod, "grammar_whitespace_mode", modgrammar.grammar_whitespace_mode)
+    cdict["grammar_whitespace_mode"] = whitespace_mode
   return cdict
 
 def calc_line_col(string, count, line=0, col=0, tabs=1):
@@ -99,21 +124,29 @@ def calc_line_col(string, count, line=0, col=0, tabs=1):
   return (line, col)
 
 def get_calling_module(stack=None):
+  return get_calling_modinfo(stack)[2]
+
+def get_calling_stacklevel(stack=None):
+  return get_calling_modinfo(stack)[0]
+
+def get_calling_modinfo(stack=None):
   if stack is None:
     stack = traceback.extract_stack(None)
+  stacklevel = -2
   for s in reversed(stack):
+    stacklevel += 1
     filename = s[0]
     if filename == "<stdin>":
-      return sys.modules["__main__"]
+      return (stacklevel, s, sys.modules["__main__"])
     elif filename == __file__ or filename == modgrammar.__file__:
       continue
     else:
       for m in sys.modules.values():
         if getattr(m, "__file__", None) == filename:
-          return m
+          return (stacklevel, s, m)
   # For some reason, we weren't able to determine the module.  Not much we
   # can do here..
-  return None
+  return (None, None, modgrammar)
 
 class RepeatingTuple (tuple):
   def __new__(cls, first_item, successive_items, len=None):
@@ -156,3 +189,39 @@ def ebnf_specialseq(grammar, opts, name=None, details=None, desc=None):
     if not text:
       text = name or grammar.grammar_name
   return '? {} ?'.format(text)
+
+def find_match_func(elem, type_or_tag):
+  try:
+    if isinstance(elem, type_or_tag):
+      return True
+  except TypeError:
+    pass
+  return type_or_tag in getattr(elem, "grammar_tags", ())
+
+def get_found_txt(buf, pos):
+  end = min(len(buf), pos + PARSEERROR_FOUNDTXT_LEN)
+  if end == pos:
+    return "(end of input)"
+  try:
+    end = min(end, buf.index('\n', pos))
+  except ValueError:
+    pass
+  try:
+    end = min(end, buf.index('\r', pos))
+  except ValueError:
+    pass
+  found_txt = buf[pos:end]
+  if found_txt:
+    return repr(found_txt)
+  else:
+    return "(end of line)"
+
+depwarnings_issued = set()
+
+def depwarning(msg, stacklevel=2, module=None):
+  if module == modgrammar:
+    return
+  winfo = (module, msg)
+  if not winfo in depwarnings_issued:
+    depwarnings_issued.add(winfo)
+    warnings.warn("modgrammar: " + msg, DeprecationWarning, stacklevel=stacklevel+1)
