@@ -10,7 +10,7 @@ from . import util
 # * collapsing
 # * collapse_skip
 
-grammar_whitespace = True
+grammar_whitespace_mode = 'optional'
 
 NoneType = type(None)
 
@@ -22,7 +22,11 @@ class G2 (Grammar):
 
 class G3 (Grammar):
   grammar = ('ABC', 'DEF')
-  grammar_whitespace = False
+  grammar_whitespace_mode = 'explicit'
+
+class G4 (Grammar):
+  grammar = ('ABC', 'DEF')
+  grammar_whitespace_mode = 'required'
 
 class TestG1 (util.BasicGrammarTestCase):
   def setUp(self):
@@ -51,8 +55,8 @@ class TestG2 (util.BasicGrammarTestCase):
     self.fail_partials = (('A', 'B', 'C', 'D', 'E', 'X'), ('ABC', 'D', 'EX'), ('AB', 'XDEF'))
 
 class TestG3 (util.BasicGrammarTestCase):
-  # For this, we're just checking that if grammar_whitespace=False, it won't
-  # match if there's whitespace between tokens.
+  # For this, we're just checking that if grammar_whitespace_mode='explicit',
+  # it won't match if there's whitespace between tokens.
   def setUp(self):
     self.grammar = G3
     self.grammar_name = "G3"
@@ -62,6 +66,18 @@ class TestG3 (util.BasicGrammarTestCase):
     self.matches = ('ABCDEF',)
     self.fail_matches = ('ABC DEF',)
 
+class TestG4 (util.BasicGrammarTestCase):
+  # For this, we're just checking that if grammar_whitespace_mode='required',
+  # it won't match if there is NOT whitespace between tokens.
+  def setUp(self):
+    self.grammar = G4
+    self.grammar_name = "G4"
+    self.grammar_details = "(L('ABC'), L('DEF'))"
+    self.terminal = False
+    self.subgrammar_types = (Literal, Literal)
+    self.matches = ('ABC DEF',)
+    self.fail_matches = ('ABCDEF',)
+
 class TestGRAMMAR (util.BasicGrammarTestCase):
   def setUp(self):
     self.grammar = GRAMMAR('ABC', 'DEF')
@@ -83,10 +99,10 @@ class TestGRAMMAR (util.BasicGrammarTestCase):
     self.assertTrue(issubclass(g, Literal))
 
 class TestGRAMMAR_NoWS (util.BasicGrammarTestCase):
-  # For this, we're just checking that if grammar_whitespace=False, it won't
-  # match if there's whitespace between tokens.
+  # For this, we're just checking that if grammar_whitespace_mode='explicit',
+  # it won't match if there's whitespace between tokens.
   def setUp(self):
-    self.grammar = GRAMMAR('ABC', 'DEF', whitespace=False)
+    self.grammar = GRAMMAR('ABC', 'DEF', whitespace_mode='explicit')
     self.grammar_name = "<GRAMMAR>"
     self.grammar_details = "(L('ABC'), L('DEF'))"
     self.terminal = False
@@ -94,6 +110,18 @@ class TestGRAMMAR_NoWS (util.BasicGrammarTestCase):
     self.matches = ('ABCDEF',)
     self.fail_matches = ('ABC DEF',)
 
+class TestGRAMMAR_WSReq (util.BasicGrammarTestCase):
+  # For this, we're just checking that if grammar_whitespace_mode='required', it
+  # won't match if there is NOT whitespace between tokens.
+  def setUp(self):
+    self.grammar = GRAMMAR('ABC', 'DEF', whitespace_mode='required')
+    self.grammar_name = "<GRAMMAR>"
+    self.grammar_details = "(L('ABC'), L('DEF'))"
+    self.terminal = False
+    self.subgrammar_types = (Literal, Literal)
+    self.matches = ('ABC DEF',)
+    self.fail_matches = ('ABCDEF',)
+
 class TestLiteral (util.BasicGrammarTestCase):
   def setUp(self):
     self.grammar = LITERAL('ABC')
@@ -193,16 +221,16 @@ class TestRepeat1 (util.BasicGrammarTestCase):
     self.assertTrue(REPEAT('ABC').grammar_greedy)
     # Test greedy behavior
     p = REPEAT('ABC', min=0, max=3, greedy=True).parser()
-    o = p.parse_string('ABCABCABCABC', matchtype='all')
+    o = p.parse_text('ABCABCABCABC', matchtype='all')
     self.assertEqual([x.string for x in o], ['ABCABCABC', 'ABCABC', 'ABC', ''])
     # Test non-greedy behavior
     p = REPEAT('ABC', min=0, max=3, greedy=False).parser()
-    o = p.parse_string('ABCABCABCABC', matchtype='all')
+    o = p.parse_text('ABCABCABCABC', matchtype='all')
     self.assertEqual([x.string for x in o], ['', 'ABC', 'ABCABC', 'ABCABCABC'])
     # Simple test of combining greedy and non-greedy repetitions
     g = GRAMMAR(REPEAT('ABC', max=3, min=0, greedy=False), REPEAT('ABC', max=3, min=0))
     p = g.parser()
-    o = p.parse_string('ABCABCABCABCx', matchtype='all')
+    o = p.parse_text('ABCABCABCABCx', matchtype='all')
     strings = [[x.string for x in y] for y in o]
     self.assertEqual(strings, [['', 'ABCABCABC'],
                                ['', 'ABCABC'],
@@ -231,7 +259,7 @@ class TestRepeat2 (util.BasicGrammarTestCase):
 
 class TestRepeat1_NoWS (util.BasicGrammarTestCase):
   def setUp(self):
-    self.grammar = REPEAT('ABC', min=2, max=5, whitespace=False)
+    self.grammar = REPEAT('ABC', min=2, max=5, whitespace_mode='explicit')
     self.grammar_name = "<REPEAT>"
     self.grammar_details = "REPEAT(L('ABC'), min=2, max=5)"
     self.subgrammar_types = (Literal, Literal, Literal, Literal, Literal)
@@ -242,6 +270,19 @@ class TestRepeat1_NoWS (util.BasicGrammarTestCase):
     self.partials = (('ABC', 'ABC', 'x'), ('ABCABC', 'x'), ('AB', 'CA', 'BC', 'x'), ('ABCABC', 'ABCABC', 'ABC'))
     self.fail_partials = (('ABC', 'ABx'), ('ABC ',), ('ABC', ' ABC'), ('ABC', ' '))
 
+class TestRepeat1_WSReq (util.BasicGrammarTestCase):
+  def setUp(self):
+    self.grammar = REPEAT('ABC', min=3, max=5, whitespace_mode='required')
+    self.grammar_name = "<REPEAT>"
+    self.grammar_details = "REPEAT(L('ABC'), min=3, max=5)"
+    self.subgrammar_types = (Literal, Literal, Literal, Literal, Literal)
+    self.terminal = False
+    self.matches = ('ABC ABC ABC ABC ABC',)
+    self.matches_with_remainder = ('ABC ABC ABCx', 'ABC ABC ABC x', 'ABC ABC ABC ABx')
+    self.fail_matches = ('ABC ABCABC', 'ABCABC ABC', 'ABC ABCx')
+    self.partials = (('ABC', ' ABC', ' ', 'ABC', 'x'), ('ABC ABC ', 'ABC', 'x'), ('AB', 'C A', 'BC', ' A', 'BC', 'x'), ('ABC ABC', ' ABC ABC', ' ABC'))
+    self.fail_partials = (('ABC', 'ABC'), ('ABC ', 'x'), ('ABC', 'A'))
+
 class TestOptional (util.BasicGrammarTestCase):
   def setUp(self):
     self.grammar = OPTIONAL('ABC')
@@ -256,7 +297,7 @@ class TestOptional (util.BasicGrammarTestCase):
 
   def test_failmatch_is_none(self):
     p = self.grammar.parser()
-    o = p.parse_string('x')
+    o = p.parse_text('x')
     self.assertEqual(p.remainder(), 'x')
     self.assertIsNone(o)
 
@@ -265,16 +306,16 @@ class TestOptional (util.BasicGrammarTestCase):
     self.assertTrue(OPTIONAL('ABC').grammar_greedy)
     # Test greedy behavior
     p = OPTIONAL('ABC', greedy=True).parser()
-    o = p.parse_string('ABC', matchtype='all')
+    o = p.parse_text('ABC', matchtype='all')
     self.assertEqual([x and x.string for x in o], ['ABC', None])
     # Test non-greedy behavior
     p = OPTIONAL('ABC', greedy=False).parser()
-    o = p.parse_string('ABC', matchtype='all')
+    o = p.parse_text('ABC', matchtype='all')
     self.assertEqual([x and x.string for x in o], [None, 'ABC'])
     # Simple test of combining greedy and non-greedy optionals
     g = GRAMMAR(OPTIONAL('ABC', greedy=False), OPTIONAL('ABC'))
     p = g.parser()
-    o = p.parse_string('ABCABC', matchtype='all')
+    o = p.parse_text('ABCABC', matchtype='all')
     strings = [[x and x.string for x in y] for y in o]
     self.assertEqual(strings, [[None, 'ABC'],
                                [None, None],
@@ -326,10 +367,10 @@ class TestList1 (util.BasicGrammarTestCase):
 
   def test_trailing_comma(self):
     p = self.grammar.parser()
-    p.parse_string('ABC,x')
+    p.parse_text('ABC,x')
     self.assertEqual(p.remainder(), ',x')
     p.reset()
-    p.parse_string('ABC,ABC,ABC,')
+    p.parse_text('ABC,ABC,ABC,')
     self.assertEqual(p.remainder(), ',')
 
 class TestList2 (util.BasicGrammarTestCase):
@@ -349,7 +390,7 @@ class TestList2 (util.BasicGrammarTestCase):
 
 class TestList_NoWS (util.BasicGrammarTestCase):
   def setUp(self):
-    self.grammar = LIST_OF('ABC', count=3, whitespace=False)
+    self.grammar = LIST_OF('ABC', count=3, whitespace_mode='explicit')
     self.grammar_name = "<LIST>"
     self.grammar_details = "LIST_OF(L('ABC'), sep=L(','), count=3)"
     self.subgrammar_types = (Literal, AnonGrammar, AnonGrammar)
@@ -362,6 +403,22 @@ class TestList_NoWS (util.BasicGrammarTestCase):
     numtok = len(teststr.split(','))*2 - 1
     return (numtok, numtok)
 
+class TestList_WSReq (util.BasicGrammarTestCase):
+  def setUp(self):
+    self.grammar = LIST_OF('ABC', count=3, whitespace_mode='required')
+    self.grammar_name = "<LIST>"
+    self.grammar_details = "LIST_OF(L('ABC'), sep=L(','), count=3)"
+    self.subgrammar_types = (Literal, AnonGrammar, AnonGrammar)
+    self.terminal = False
+    self.matches = ('ABC , ABC , ABC',)
+    self.fail_matches = ('ABC,ABC , ABC', 'ABC, ABC , ABC', 'ABC ,ABC , ABC')
+    self.partials = (('ABC', ' ', ',', ' ', 'ABC , ABC'), ('A', 'B', 'C', ' , ', 'ABC ', ',', ' ABC'))
+    self.fail_partials = (('ABC', ','), ('ABC', ' ,', 'ABC'), ('ABC', ' ', ',', 'ABC'))
+
+  def num_tokens_for(self, teststr):
+    numtok = len(teststr.split(','))*2 - 1
+    return (numtok, numtok)
+
 class TestWord1 (util.BasicGrammarTestCase):
   def setUp(self):
     self.grammar = WORD('a-z', min=2, max=5)
@@ -378,38 +435,38 @@ class TestWord1 (util.BasicGrammarTestCase):
     g = GRAMMAR(self.grammar, 'a')
     p = g.parser()
     with self.assertRaises(ParseError):
-      p.parse_string('abcdefa') # doesn't have 'a' after the max word length
+      p.parse_text('abcdefa') # doesn't have 'a' after the max word length
     p.reset()
-    o = p.parse_string('abcdea')
+    o = p.parse_text('abcdea')
     self.assertEqual(o.tokens(), ['abcde', 'a'])
     p.reset()
-    o = p.parse_string('abcdae')
+    o = p.parse_text('abcdae')
     self.assertEqual(o.tokens(), ['abcd', 'a'])
     p.reset()
-    o = p.parse_string('abcade')
+    o = p.parse_text('abcade')
     self.assertEqual(o.tokens(), ['abc', 'a'])
     p.reset()
-    o = p.parse_string('abacde')
+    o = p.parse_text('abacde')
     self.assertEqual(o.tokens(), ['ab', 'a'])
     p.reset()
     with self.assertRaises(ParseError):
-      p.parse_string('aabcde') # has to backtrack too far
+      p.parse_text('aabcde') # has to backtrack too far
 
   def test_greedy(self):
     # Make sure the default is to be greedy
     self.assertTrue(WORD('a-z').grammar_greedy)
     # Test greedy behavior
     p = WORD('a-z', min=0, max=3, greedy=True).parser()
-    o = p.parse_string('abcd', matchtype='all')
+    o = p.parse_text('abcd', matchtype='all')
     self.assertEqual([x.string for x in o], ['abc', 'ab', 'a', ''])
     # Test non-greedy behavior
     p = WORD('a-z', min=0, max=3, greedy=False).parser()
-    o = p.parse_string('abcd', matchtype='all')
+    o = p.parse_text('abcd', matchtype='all')
     self.assertEqual([x.string for x in o], ['', 'a', 'ab', 'abc'])
     # Simple test of combining greedy and non-greedy words
     g = GRAMMAR(WORD('a-z', max=3, min=0, greedy=False), WORD('a-z', max=3, min=0))
     p = g.parser()
-    o = p.parse_string('abcdefg', matchtype='all')
+    o = p.parse_text('abcdefg', matchtype='all')
     strings = [[x.string for x in y] for y in o]
     self.assertEqual(strings, [['', 'abc'],
                                ['', 'ab'],
@@ -443,13 +500,37 @@ class TestWord2 (util.BasicGrammarTestCase):
   def test_backtrack(self):
     g = GRAMMAR(self.grammar, 'a')
     p = g.parser()
-    o = p.parse_string('Abcade')
+    o = p.parse_text('Abcade')
     self.assertEqual(o.tokens(), ['Abc', 'a'])
     g = GRAMMAR(self.grammar, 'A')
     p = g.parser()
-    o = p.parse_string('Abcade')
+    o = p.parse_text('Abcade')
     self.assertEqual(o.tokens(), ['', 'A'])
 
+class TestWord3 (util.BasicGrammarTestCase):
+  def setUp(self):
+    self.grammar = WORD('a-z', min=2, max=5, longest=True)
+    self.grammar_name = "WORD('a-z')"
+    self.grammar_details = "WORD('a-z', min=2, max=5, longest=True)"
+    self.matches = ('abcde',)
+    self.matches_with_remainder = ('abcdef', 'ab#cd')
+    self.fail_matches = ('Abcd', '\nabcde', 'a bcd')
+    self.partials = (('a', 'b', 'c', 'd', 'e'), ('a', 'b', '#'))
+    self.fail_partials = (('a', '#'),)
+
+  # There should be no backtracking on longest=True WORD grammars
+  def test_backtrack(self):
+    g = GRAMMAR(self.grammar, 'a')
+    p = g.parser()
+    o = p.parse_text('abcdea')
+    self.assertEqual(o.tokens(), ['abcde', 'a'])
+    p.reset()
+    o = p.parse_text('abcd a')
+    self.assertEqual(o.tokens(), ['abcd', 'a'])
+    p.reset()
+    with self.assertRaises(ParseError):
+      p.parse_text('abcdab') # Would be OK except for longest=True
+
 class TestAnyExcept (util.BasicGrammarTestCase):
   def setUp(self):
     self.grammar = ANY_EXCEPT('a-z')
@@ -461,9 +542,9 @@ class TestAnyExcept (util.BasicGrammarTestCase):
 
   def test_pre_post_space(self):
     # This would normally fail the default pre/post space tests, because it has
-    # grammar_whitespace=False, but still matches fine if there's whitespace at
-    # the beginning (because whitespace characters match the ANY_EXCEPT
-    # criteria), so we'll just skip that test in this case.
+    # grammar_whitespace_mode='explicit', but still matches fine if there's
+    # whitespace at the beginning (because whitespace characters match the
+    # ANY_EXCEPT criteria), so we'll just skip that test in this case.
     pass
 
 class TestBOL (util.BasicGrammarTestCase):
@@ -476,21 +557,21 @@ class TestBOL (util.BasicGrammarTestCase):
 
   def test_match_fail(self):
     with self.assertRaises(ParseError):
-      self.grammar.parser().parse_string('a', bol=False)
+      self.grammar.parser().parse_text('a', bol=False)
 
   def test_mid_string(self):
     grammar = GRAMMAR(ANY, BOL)
     p = grammar.parser()
-    o = p.parse_string('a\na')
+    o = p.parse_text('a\na')
     self.assertIsNotNone(o)
     self.assertEqual(p.remainder(), 'a')
-    grammar = GRAMMAR(ANY, ANY, BOL, whitespace=False)
+    grammar = GRAMMAR(ANY, ANY, BOL, whitespace_mode='explicit')
     p = grammar.parser()
-    o = p.parse_string('a\na')
+    o = p.parse_text('a\na')
     self.assertIsNotNone(o)
     self.assertEqual(p.remainder(), 'a')
     p.reset()
-    o = p.parse_string('a\r')
+    o = p.parse_text('a\r')
     self.assertIsNotNone(o)
     self.assertEqual(p.remainder(), '')
 
@@ -516,9 +597,9 @@ class TestRestOfLine (util.BasicGrammarTestCase):
 
   def test_pre_post_space(self):
     # This would normally fail the default pre/post space tests, because it has
-    # grammar_whitespace=False, but still matches fine if there's whitespace at
-    # the beginning (because whitespace characters match the REST_OF_LINE
-    # criteria), so we'll just skip that test in this case.
+    # grammar_whitespace_mode='explicit', but still matches fine if there's
+    # whitespace at the beginning (because whitespace characters match the
+    # REST_OF_LINE criteria), so we'll just skip that test in this case.
     pass
 
 class TestANY (util.BasicGrammarTestCase):
@@ -531,27 +612,27 @@ class TestANY (util.BasicGrammarTestCase):
 
   def test_pre_post_space(self):
     # This would normally fail the default pre/post space tests, because it has
-    # grammar_whitespace=False, but still matches if there's whitespace at
-    # the beginning (because whitespace characters match the ANY
-    # criteria), so we'll just skip that test in this case.
+    # grammar_whitespace_mode='explicit', but still matches fine if there's
+    # whitespace at the beginning (because whitespace characters match the
+    # ANY criteria), so we'll just skip that test in this case.
     # Actual testing of whitespace before/after a string is already handled by
     # the matches_with_remainder test cases.
     pass
 
 class TestSpace (util.BasicGrammarTestCase):
   def setUp(self):
-    self.grammar = SPACE
-    self.grammar_name = "SPACE"
-    self.grammar_details = "SPACE"
+    self.grammar = WHITESPACE
+    self.grammar_name = "WHITESPACE"
+    self.grammar_details = "WHITESPACE"
     self.matches_with_remainder = (' \t\r\na', '   a')
     self.fail_matches = ('a', 'a ')
     self.partials = ((' ', 'a'),)
 
   def test_pre_post_space(self):
     # This would normally fail the default pre/post space tests, because it has
-    # grammar_whitespace=False, but still matches fine if there's whitespace at
-    # the beginning (because whitespace characters match the SPACE
-    # criteria), so we'll just skip that test in this case.
+    # grammar_whitespace_mode='explicit', but still matches fine if there's
+    # whitespace at the beginning (because whitespace characters match the
+    # WHITESPACE criteria), so we'll just skip that test in this case.
     pass
 
 class TestEOF (util.BasicGrammarTestCase):
@@ -566,7 +647,7 @@ class TestEOF (util.BasicGrammarTestCase):
     g = self.grammar
     p = g.parser()
     try:
-      o = p.parse_string('', eof=True)
+      o = p.parse_text('', eof=True)
       self.check_result('', o, True, '')
     except ParseError:
       self.fail("Got unexpected ParseError")
@@ -577,15 +658,15 @@ class TestEOF (util.BasicGrammarTestCase):
       p.reset()
       msg = '[testcase={!r}]'.format(teststr)
       with self.assertRaises(ParseError, msg=msg):
-        p.parse_string(teststr, eof=True)
+        p.parse_text(teststr, eof=True)
 
   def test_partial(self):
     g = self.grammar
     p = g.parser()
     try:
-      o = p.parse_string('')
+      o = p.parse_text('')
       self.assertIsNone(o)
-      o = p.parse_string('', eof=True)
+      o = p.parse_text('', eof=True)
       self.check_result('', o, True, '')
     except ParseError:
       self.fail("Got unexpected ParseError")
@@ -596,7 +677,7 @@ class TestEOF (util.BasicGrammarTestCase):
     try:
       for teststr in ('', ' '):
         msg = '[testcase={!r}]'.format(teststr)
-        o = p.parse_string(teststr, eof=True)
+        o = p.parse_text(teststr, eof=True)
         self.assertIsNotNone(o)
         self.assertEqual(len(p.remainder()), 0)
     except ParseError:
