@@ -2,6 +2,8 @@ import modgrammar
 from modgrammar import *
 from . import util
 
+grammar_whitespace_mode = 'optional'
+
 class TestIssue1 (util.TestCase):
   """
   Issue 1: Fails with traceback if matchtype="longest" or "shortest" and
@@ -10,11 +12,11 @@ class TestIssue1 (util.TestCase):
 
   def test_longest_samelength(self):
     grammar = OR('aaa', 'aaa')
-    o = grammar.parser().parse_string('aaaa', matchtype='longest')
+    o = grammar.parser().parse_text('aaaa', matchtype='longest')
 
   def test_shortest_samelength(self):
     grammar = OR('aaa', 'aaa')
-    o = grammar.parser().parse_string('aaaa', matchtype='shortest')
+    o = grammar.parser().parse_text('aaaa', matchtype='shortest')
 
 class TestIssue2 (util.TestCase):
   """
@@ -33,6 +35,28 @@ class TestIssue4 (util.TestCase):
   """
 
   def test_whitespace_before_eof(self):
-    grammar = GRAMMAR('A', whitespace=True)
+    grammar = GRAMMAR('A', whitespace_mode='optional')
     list(grammar.parser().parse_lines(["A "], eof=True))
 
+class TestIssue10 (util.TestCase):
+  """
+  Issue 10: EXCEPT incorrectly matches on prefixes of the result
+  """
+
+  def test_except_prefix(self):
+    grammar = EXCEPT('foobar', 'foo')
+    try:
+      o = grammar.parser().parse_text('foobar')
+    except ParseError:
+      self.fail("EXCEPT incorrectly matching on substring")
+
+class TestIssue8 (util.TestCase):
+  """
+  Issue 8: Grammars with NOT_FOLLOWED_BY don't correctly return a parse error
+           before terminating the generator
+  """
+
+  def test_iteration_not_followed_by(self):
+    grammar = G('A', NOT_FOLLOWED_BY('A'))
+    o = grammar.parser().parse_text('AB', matchtype='all')
+    self.assertEqual([x.string for x in o], ['A'])
